# This file contains the list of providers and models that are available supported by LLMX.


from llmx.utils import load_config


config = load_config()
providers = providers = config["providers"] if "providers" in config else None

providers = config["providers"]



from dataclasses import asdict
import logging
import json
from typing import Any, Union, Dict
import tiktoken
from diskcache import Cache
import hashlib
import os
import platform
import google.auth
import google.auth.transport.requests
from google.oauth2 import service_account
import requests
import yaml

logger = logging.getLogger("llmx")


def num_tokens_from_messages(messages, model="gpt-3.5-turbo-0301"):
    """Returns the number of tokens used by a list of messages."""
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        encoding = tiktoken.get_encoding("cl100k_base")
    if (
        model == "gpt-3.5-turbo-0301" or True
    ):  # note: future models may deviate from this
        num_tokens = 0
        for message in messages:
            if not isinstance(message, dict):
                message = asdict(message)

            num_tokens += (
                4  # every message follows <im_start>{role/name}\n{content}<im_end>\n
            )

            for key, value in message.items():
                num_tokens += len(encoding.encode(value))
                if key == "name":  # if there's a name, the role is omitted
                    num_tokens += -1  # role is always required and always 1 token
        num_tokens += 2  # every reply is primed with <im_start>assistant
        return num_tokens


def cache_request(cache: Cache, params: dict, values: Union[Dict, None] = None) -> Any:
    # Generate a unique key for the request

    key = hashlib.md5(json.dumps(params, sort_keys=True).encode("utf-8")).hexdigest()
    # Check if the request is cached
    if key in cache and values is None:
        # print("retrieving from cache")
        return cache[key]

    # Cache the provided values and return them
    if values:
        # print("saving to cache")
        cache[key] = values
    return values


def get_user_cache_dir(app_name: str) -> str:
    system = platform.system()
    if system == "Windows":
        cache_path = os.path.join(os.getenv("LOCALAPPDATA"), app_name, "Cache")
    elif system == "Darwin":
        cache_path = os.path.join(os.path.expanduser("~/Library/Caches"), app_name)
    else:  # Linux and other UNIX-like systems
        cache_path = os.path.join(
            os.getenv("XDG_CACHE_HOME", os.path.expanduser("~/.cache")), app_name
        )
    os.makedirs(cache_path, exist_ok=True)
    return cache_path


def get_gcp_credentials(service_account_key_file: str = None, scopes: list[str] = [
        'https://www.googleapis.com/auth/cloud-platform']):
    try:
        # Attempt to use Application Default Credentials
        credentials, project_id = google.auth.default(scopes=scopes)
        auth_req = google.auth.transport.requests.Request()
        credentials.refresh(auth_req)
        return credentials
    except google.auth.exceptions.DefaultCredentialsError:
        # Fall back to using service account key
        if service_account_key_file is None:
            raise ValueError(
                "Service account key file is not set. Please set the PALM_SERVICE_ACCOUNT_KEY_FILE environment variable."
            )
        credentials = service_account.Credentials.from_service_account_file(
            service_account_key_file, scopes=scopes)
        auth_req = google.auth.transport.requests.Request()
        credentials.refresh(auth_req)
        return credentials


def gcp_request(
    url: str,
    method: str = "POST",
    body: dict = None,
    headers: dict = None,
    credentials: google.auth.credentials.Credentials = None,
    request_timeout: int = 60,
    **kwargs,
):

    headers = headers or {}

    if "key" not in url:
        if credentials is None:
            credentials = get_gcp_credentials()
        auth_req = google.auth.transport.requests.Request()
        if credentials.expired:
            credentials.refresh(auth_req)
        headers["Authorization"] = f"Bearer {credentials.token}"
    headers["Content-Type"] = "application/json"

    response = requests.request(
        method=method, url=url, json=body, headers=headers, timeout=request_timeout, **kwargs
    )

    if response.status_code not in range(200, 300):
        try:
            error_message = response.json().get("error", {}).get("message", "")
        except json.JSONDecodeError:
            error_message = response.content
        raise Exception(
            f"Request failed with status code {response.status_code}: {error_message}"
        )

    return response.json()


def load_config():
    try:
        config_path = os.environ.get("LLMX_CONFIG_PATH", None)
        if config_path is None or os.path.exists(config_path) is False:
            config_path = os.path.join(
                os.path.dirname(__file__),
                "configs/config.default.yml")
            logger.info(
                "Info: LLMX_CONFIG_PATH environment variable is not set to a valid config file. Using default config file at '%s'.",
                config_path)
        if config_path is not None:
            try:
                with open(config_path, "r", encoding="utf-8") as f:
                    config = yaml.safe_load(f)
                    logger.info(
                        "Loaded config from '%s'.",
                        config_path)
                    return config
            except FileNotFoundError as file_not_found:
                logger.info(
                    "Error: Config file not found at '%s'. Please check the LLMX_CONFIG_PATH environment variable. %s",
                    config_path,
                    str(file_not_found))
            except IOError as io_error:
                logger.info(
                    "Error: Could not read the config file at '%s'. %s",
                    config_path, str(io_error))
            except yaml.YAMLError as yaml_error:
                logger.info(
                    "Error: Malformed YAML in config file at '%s'. %s",
                    config_path, str(yaml_error))
        else:
            logger.info(
                "Info:LLMX_CONFIG_PATH environment variable is not set. Please set it to the path of your config file to setup your default model.")
    except Exception as error:
        logger.info("Error: An unexpected error occurred: %s", str(error))

    return None


def get_models_maxtoken_dict(models_list):
    models_dict = {}
    for model in models_list:
        if "model" in model and "parameters" in model["model"]:
            details = model["model"]["parameters"]
            models_dict[details["model"]] = model["max_tokens"]
    return models_dict



import sys
from .generators.text.textgen import llm
from .datamodel import TextGenerationConfig, TextGenerationResponse, Message
from .generators.text.base_textgen import TextGenerator
from .generators.text.providers import providers

if sys.version_info < (3, 9):
    raise RuntimeError("llmx requires Python 3.10+")



from ...utils import load_config
from .openai_textgen import OpenAITextGenerator
from .palm_textgen import PalmTextGenerator
from .cohere_textgen import CohereTextGenerator
import logging

logger = logging.getLogger("llmx")


def sanitize_provider(provider: str):
    if provider.lower() == "openai" or provider.lower() == "default" or provider.lower(
    ) == "azureopenai" or provider.lower() == "azureoai":
        return "openai"
    elif provider.lower() == "palm" or provider.lower() == "google":
        return "palm"
    elif provider.lower() == "cohere":
        return "cohere"
    elif provider.lower() == "hf" or provider.lower() == "huggingface":
        return "hf"
    else:
        raise ValueError(
            f"Invalid provider '{provider}'.  Supported providers are 'openai', 'hf', 'palm', and 'cohere'."
        )


def llm(provider: str = None, **kwargs):

    # load config. This will load the default config from
    # configs/config.default.yml if no path to a config file is specified. in
    # the environment variable LLMX_CONFIG_PATH
    config = load_config()
    if provider is None:
        # provider is not explicitly specified, use the default provider from the config file
        provider = config["model"]["provider"] if "provider" in config["model"] else None
        kwargs = config["model"]["parameters"] if "parameters" in config["model"] else {}
    if provider is None:
        logger.info("No provider specified. Defaulting to 'openai'.")
        provider = "openai"

    # sanitize provider
    provider = sanitize_provider(provider)

    # set the list of available models based on the config file
    models = config["providers"][provider]["models"] if "providers" in config and provider in config["providers"] else {}

    # print(kwargs)

    if provider.lower() == "openai":
        return OpenAITextGenerator(models=models, **kwargs)
    elif provider.lower() == "palm":
        return PalmTextGenerator(provider=provider, models=models, **kwargs)
    elif provider.lower() == "cohere":
        return CohereTextGenerator(provider=provider, models=models, **kwargs)
    elif provider.lower() == "hf":
        try:
            import transformers
        except ImportError:
            raise ImportError(
                "Please install the `transformers` package to use the HFTextGenerator class. pip install llmx[transformers]"
            )

        # Check if torch package is installed
        try:
            import torch
        except ImportError:
            raise ImportError(
                "Please install the `torch` package to use the HFTextGenerator class.  pip install llmx[transformers]"
            )

        from .hf_textgen import HFTextGenerator

        return HFTextGenerator(provider=provider, models=models, **kwargs)

    else:
        raise ValueError(
            f"Invalid provider '{provider}'.  Supported providers are 'openai', 'hf', 'palm', and 'cohere'."
        )



from dataclasses import asdict
import os
import logging
from typing import Dict, Union
from .base_textgen import TextGenerator
from ...datamodel import TextGenerationConfig, TextGenerationResponse, Message
from ...utils import (
    cache_request,
    gcp_request,
    get_models_maxtoken_dict,
    num_tokens_from_messages,
    get_gcp_credentials,
)

logger = logging.getLogger("llmx")


class PalmTextGenerator(TextGenerator):
    def __init__(
        self,
        api_key: str = os.environ.get("PALM_API_KEY", None),
        palm_key_file: str = os.environ.get("PALM_SERVICE_ACCOUNT_KEY_FILE", None),
        project_id: str = os.environ.get("PALM_PROJECT_ID", None),
        project_location=os.environ.get("PALM_PROJECT_LOCATION", "us-central1"),
        provider: str = "palm",
        model: str = None,
        models: Dict = None,
    ):
        super().__init__(provider=provider)

        if api_key is None and palm_key_file is None:
            raise ValueError(
                "PALM_API_KEY or PALM_SERVICE_ACCOUNT_KEY_FILE  must be set."
            )
        if api_key:
            self.api_key = api_key
            self.credentials = None
            self.project_id = None
            self.project_location = None
        else:
            self.project_id = project_id
            self.project_location = project_location
            self.api_key = None
            self.credentials = get_gcp_credentials(palm_key_file) if palm_key_file else None

        self.model_max_token_dict = get_models_maxtoken_dict(models)
        self.model_name = model or "chat-bison"

    def format_messages(self, messages):
        palm_messages = []
        system_messages = ""
        for message in messages:
            if message["role"] == "system":
                system_messages += message["content"] + "\n"
            else:
                if not palm_messages or palm_messages[-1]["author"] != message["role"]:
                    palm_message = {
                        "author": message["role"],
                        "content": message["content"],
                    }
                    palm_messages.append(palm_message)
                else:
                    palm_messages[-1]["content"] += "\n" + message["content"]

        if palm_messages and len(palm_messages) % 2 == 0:
            merged_content = (
                palm_messages[-2]["content"] + "\n" + palm_messages[-1]["content"]
            )
            palm_messages[-2]["content"] = merged_content
            palm_messages.pop()

        if len(palm_messages) == 0:
            logger.info("No messages to send to PALM")

        return system_messages, palm_messages

    def generate(
        self,
        messages: Union[list[dict], str],
        config: TextGenerationConfig = TextGenerationConfig(),
        **kwargs,
    ) -> TextGenerationResponse:
        use_cache = config.use_cache
        model = config.model or self.model_name

        system_messages, messages = self.format_messages(messages)
        self.model_name = model

        max_tokens = self.model_max_token_dict[model] if model in self.model_max_token_dict else 1024
        palm_config = {
            "temperature": config.temperature,
            "maxOutputTokens": config.max_tokens or max_tokens,
            "candidateCount": config.n,
        }

        api_url = ""
        if self.api_key:
            api_url = f"https://generativelanguage.googleapis.com/v1beta2/models/{model}:generateMessage?key={self.api_key}"

            palm_payload = {
                "prompt": {"messages": messages, "context": system_messages},
                "temperature": config.temperature,
                "candidateCount": config.n,
                "topP": config.top_p,
                "topK": config.top_k,
            }

        else:
            api_url = f"https://us-central1-aiplatform.googleapis.com/v1/projects/{self.project_id}/locations/{self.project_location}/publishers/google/models/{model}:predict"

            palm_payload = {
                "instances": [
                    {
                        "messages": messages,
                        "context": system_messages,
                        "examples": [],
                    }
                ],
                "parameters": palm_config,
            }

        cache_key_params = {**palm_payload, "model": model, "api_url": api_url}

        if use_cache:
            response = cache_request(cache=self.cache, params=cache_key_params)
            if response:
                return TextGenerationResponse(**response)

        palm_response = gcp_request(
            url=api_url, body=palm_payload, method="POST", credentials=self.credentials
        )

        candidates = palm_response["candidates"] if self.api_key else palm_response["predictions"][
            0]["candidates"]

        response_text = [
            Message(
                role="assistant" if x["author"] == "1" else x["author"],
                content=x["content"],
            )
            for x in candidates
        ]

        response = TextGenerationResponse(
            text=response_text,
            logprobs=[],
            config=palm_config,
            usage={
                "total_tokens": num_tokens_from_messages(
                    response_text, model=self.model_name
                )
            },
            response=palm_response,
        )

        cache_request(
            cache=self.cache, params=(cache_key_params), values=asdict(response)
        )
        return response

    def count_tokens(self, text) -> int:
        return num_tokens_from_messages(text)



from typing import Union, List, Dict
from .base_textgen import TextGenerator
from ...datamodel import Message, TextGenerationConfig, TextGenerationResponse
from ...utils import cache_request, get_models_maxtoken_dict, num_tokens_from_messages
import os
import openai
from dataclasses import asdict


class OpenAITextGenerator(TextGenerator):
    def __init__(
        self,
        api_key: str = os.environ.get("OPENAI_API_KEY", None),
        provider: str = "openai",
        organization: str = None,
        api_type: str = None,
        api_base: str = None,
        api_version: str = None,
        model: str = None,
        models: Dict = None,
    ):
        super().__init__(provider=provider)
        api_key = api_key or os.environ.get("OPENAI_API_KEY", None)

        if api_key is None:
            raise ValueError(
                "OpenAI API key is not set. Please set the OPENAI_API_KEY environment variable."
            )
        openai.api_key = api_key
        if organization:
            openai.organization = organization
        if api_version:
            openai.api_version = api_version
        if api_base:
            openai.api_base = api_base
        if api_type:
            openai.api_type = api_type

        self.model_name = model or "gpt-3.5-turbo"

        self.model_max_token_dict = get_models_maxtoken_dict(models)
        # print("context lengths", self.model_max_token_dict)

    def generate(
        self,
        messages: Union[List[dict], str],
        config: TextGenerationConfig = TextGenerationConfig(),
        **kwargs,
    ) -> TextGenerationResponse:
        use_cache = config.use_cache
        model = config.model or self.model_name
        prompt_tokens = num_tokens_from_messages(messages)
        max_tokens = max(self.model_max_token_dict.get(model, 4096) - prompt_tokens - 10, 200)

        oai_config = {
            "model": model,
            "temperature": config.temperature,
            "max_tokens": max_tokens,
            "top_p": config.top_p,
            "frequency_penalty": config.frequency_penalty,
            "presence_penalty": config.presence_penalty,
            "n": config.n,
            "messages": messages,
        }

        if openai.api_type and openai.api_type == "azure":
            oai_config["engine"] = config.model

        self.model_name = model
        cache_key_params = (oai_config) | {"messages": messages}
        if use_cache:
            response = cache_request(cache=self.cache, params=cache_key_params)
            if response:
                return TextGenerationResponse(**response)

        oai_response = openai.ChatCompletion.create(**oai_config)

        response = TextGenerationResponse(
            text=[Message(**x.message) for x in oai_response.choices],
            logprobs=[],
            config=oai_config,
            usage=dict(oai_response.usage),
        )
        # if use_cache:
        cache_request(
            cache=self.cache, params=cache_key_params, values=asdict(response)
        )
        return response

    def count_tokens(self, text) -> int:
        return num_tokens_from_messages(text)



from typing import Dict, Union
import os
import cohere
from dataclasses import asdict

from .base_textgen import TextGenerator
from ...datamodel import TextGenerationConfig, TextGenerationResponse, Message
from ...utils import cache_request, get_models_maxtoken_dict, num_tokens_from_messages
from ..text.providers import providers


class CohereTextGenerator(TextGenerator):
    def __init__(
        self,
        api_key: str = None,
        provider: str = "cohere",
        model: str = None,
        models: Dict = None,
    ):
        super().__init__(provider=provider)
        api_key = api_key or os.environ.get("COHERE_API_KEY", None)
        if api_key is None:
            raise ValueError(
                "Cohere API key is not set. Please set the COHERE_API_KEY environment variable."
            )
        self.client = cohere.Client(api_key)
        self.model_max_token_dict = get_models_maxtoken_dict(models)
        self.model_name = model or "command"

    def format_messages(self, messages):
        prompt = ""
        for message in messages:
            if message["role"] == "system":
                prompt += message["content"] + "\n"
            else:
                prompt += message["role"] + ": " + message["content"] + "\n"

        return prompt

    def generate(
        self,
        messages: Union[list[dict], str],
        config: TextGenerationConfig = TextGenerationConfig(),
        **kwargs,
    ) -> TextGenerationResponse:
        use_cache = config.use_cache
        messages = self.format_messages(messages)
        self.model_name = config.model or self.model_name

        max_tokens = (
            self.model_max_token_dict[self.model_name]
            if config.model in self.model_max_token_dict else 1024)

        cohere_config = {
            "model": self.model_name,
            "prompt": messages,
            "max_tokens": config.max_tokens or max_tokens,
            "temperature": config.temperature,
            "k": config.top_k,
            "p": config.top_p,
            "num_generations": config.n,
            "stop_sequences": config.stop,
            "frequency_penalty": config.frequency_penalty,
            "presence_penalty": config.presence_penalty,
        }

        # print("calling cohere ***************", config)

        cache_key_params = cohere_config | {"messages": messages}
        if use_cache:
            response = cache_request(cache=self.cache, params=cache_key_params)
            if response:
                return TextGenerationResponse(**response)

        co_response = self.client.generate(**cohere_config)

        response_text = [
            Message(
                role="system",
                content=x.text,
            )
            for x in co_response.generations
        ]

        response = TextGenerationResponse(
            text=response_text,
            logprobs=[],  # You may need to extract log probabilities from the response if needed
            config=cohere_config,
            usage={},
            response=co_response,
        )

        cache_request(
            cache=self.cache, params=cache_key_params, values=asdict(response)
        )
        return response

    def count_tokens(self, text) -> int:
        return num_tokens_from_messages(text)
